/******************************************************************************
 * pcm.c
 *-----------------------------------------------------------------------------
 * Copyright (C) 2014-2015 Osman Celimli
 * For conditions of distribution and use, see copyright notice in hscc.c
 ******************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "bytes.h"
#include "hu.h"
#include "pcm.h"
/******************************************************************************
 * Text Tables
 ******************************************************************************/
char REM_PCMMode[]=		"PCM Mode...\n";
char REM_PCMDone[]=		"The current PCM binary is : ";

char ERR_NoPCMMem[]=	"ERROR- Out of memory\n";
char ERR_PCMEarlyEnd[]=	"ERROR- Unexpected end of include file\n";
char ERR_PCMBadWrite[]=	"ERROR- Couldn't write to output file\n";

char ERR_PCMTooMuch[]=	"ERROR- Too many PCM Streams (max is ";
char ERR_PCMTooLong[]=	"ERROR- PCM Stream too large (max is ";
/******************************************************************************
 * int pcm_buildBlock(FILE *pDest, FILE **ppSrcs, unsigned int uiFiles)
 * Attempt to build a PCM binary block using uiFiles sources in **ppSrcs.
 * The final output will be written to *pDest. All files must already be open.
 ******************************************************************************/
int pcm_buildBlock(FILE *pDest, FILE **ppSrcs, unsigned int uiFiles)
{
	PCM_Entry *pDir;
	unsigned char ucFiles;
	unsigned int uiDirBase,uiFileBase;
	unsigned int uiTotalSize;

	FILE *pCurFile;
	unsigned char *pPack;
	unsigned int uiReadSize,uiWriteSize;
	unsigned int uiCurSize,uiCurOffset;
	int8p8_t SubLen;
	double dDur;

	int iSuccess;
	unsigned int uiX,uiY;

	/**
	 * Check to make sure our number of pcm includes is within a
	 * sane range, then allocate and initialize our directory table.
	 */
	pPack = NULL;
	iSuccess = -1;
	fprintf(stdout, "%s", REM_PCMMode);
	if(PCM_MAX_ENTRIES < uiFiles)
	{
		fprintf(stderr, "%s%u)\n", ERR_PCMTooMuch, PCM_MAX_ENTRIES);
		return iSuccess;
	}
	if(NULL == (pDir = malloc(sizeof(PCM_Entry) * uiFiles)))
	{
		fprintf(stderr, "%s", ERR_NoPCMMem);
		return iSuccess;
	}
	for(uiX = 0; uiX < uiFiles; uiX++)
	{
		pDir[uiX].pSampleData = NULL;
		pDir[uiX].Size.uiWhole = 0;
	}


	/**
	 * Start building the directory table.
	 * Check the size and offset of each included file, then
	 * load them all in for the building phase. Don't forget we
	 * also need to shift the 8-Bit unsigned sample data over
	 * for the final 5-Bit writes.
	 */
	uiDirBase = PCM_HEADER_LEN;
	uiFileBase = (PCM_ENTRY_LEN * uiFiles) + uiDirBase;
	uiCurOffset = uiFileBase;
	for(uiX = 0; uiX < uiFiles; uiX++)
	{
		pCurFile = ppSrcs[uiX];

		fseek(pCurFile, 0, SEEK_END);
		uiCurSize = ftell(pCurFile);
		rewind(pCurFile);

		if(PCM_MAX_STREAMLEN < uiCurSize)
		{
			fprintf(stderr, "%s%u)\n", ERR_PCMTooLong, PCM_MAX_STREAMLEN);
			goto cleanup;
		}

		pDir[uiX].HuOffset = hu_calcAddr(uiCurOffset);
		if(NULL == (pDir[uiX].pSampleData = malloc(sizeof(char) * uiCurSize)))
		{
			fprintf(stderr, "%s", ERR_NoPCMMem);
			goto cleanup;
		}
		pDir[uiX].Size.uiWhole = uiCurSize;
		uiReadSize = fread(pDir[uiX].pSampleData, sizeof(char), uiCurSize, pCurFile);
		if(uiReadSize != uiCurSize)
		{
			fprintf(stderr, "%s", ERR_PCMEarlyEnd);
			goto cleanup;
		}

		for(uiY = 0; uiY < uiCurSize; uiY++)
		{
			pDir[uiX].pSampleData[uiY] = (pDir[uiX].pSampleData[uiY] >> 3);
		}
		uiCurOffset += uiCurSize;
	}
	uiTotalSize = uiCurOffset;


	/**
	 * All of our source data are ready, so let's build the
	 * binary. First stamp the header, then the directory
	 * entries, and finally the PCM streams.
	 */
	if(NULL == (pPack = malloc(sizeof(char) * uiTotalSize)))
	{
		fprintf(stderr, "%s", ERR_NoPCMMem);
		goto cleanup;
	}

	strcpy((char*)pPack + PCM_HEADER_OFFSET_MAGIC, PCM_MAGICWORD);
	ucFiles = uiFiles;
	pPack[PCM_HEADER_OFFSET_COUNT] = ucFiles;

	uiCurOffset = PCM_HEADER_LEN;
	for(uiX = 0; uiX < uiFiles; uiX++)
	{
		pPack[uiCurOffset + PCM_ENTRY_OFFSET_ADDRL] = pDir[uiX].HuOffset.ucLow;
		pPack[uiCurOffset + PCM_ENTRY_OFFSET_ADDRH] = pDir[uiX].HuOffset.ucHigh;
		pPack[uiCurOffset + PCM_ENTRY_OFFSET_ADDRB] = pDir[uiX].HuOffset.ucBank;
		pPack[uiCurOffset + PCM_ENTRY_OFFSET_ADDRM] = pDir[uiX].HuOffset.ucMap;

		SubLen.usWhole = pDir[uiX].Size.usPair.usHi;
		pPack[uiCurOffset + PCM_ENTRY_OFFSET_SIZEH] = SubLen.ucPair.ucLo;
		SubLen.usWhole = pDir[uiX].Size.usPair.usLo;
		pPack[uiCurOffset + PCM_ENTRY_OFFSET_SIZEM] = SubLen.ucPair.ucHi;
		pPack[uiCurOffset + PCM_ENTRY_OFFSET_SIZEL] = SubLen.ucPair.ucLo;

		pPack[uiCurOffset + PCM_ENTRY_OFFSET_MAGIC] = PCM_ENTRY_MAGICWORD;

		uiCurOffset += PCM_ENTRY_LEN;
	}

	uiCurOffset = uiFileBase;
	for(uiX = 0; uiX < uiFiles; uiX++)
	{
		memcpy((pPack + uiCurOffset), pDir[uiX].pSampleData,
			pDir[uiX].Size.uiWhole);
		uiCurOffset += pDir[uiX].Size.uiWhole;
	}
	
	uiWriteSize = fwrite(pPack, sizeof(char), uiTotalSize, pDest);
	if(uiWriteSize != uiTotalSize)
	{
		fprintf(stderr, "%s", ERR_PCMBadWrite);
		goto cleanup;
	}

	dDur = uiTotalSize;
	dDur /= PCM_PLAYBACK_RATE;
	fprintf(stdout, "%s%u bytes / %lf seconds\n",
		REM_PCMDone, uiTotalSize, dDur);
	iSuccess = 0;

cleanup:
	if(NULL != pPack) free(pPack);
	if(NULL != pDir)
	{
		for(uiX = 0; uiX < uiFiles; uiX++)
		{
			if(NULL != pDir[uiX].pSampleData) free(pDir[uiX].pSampleData);
		}
		free(pDir);
	}
	return iSuccess;
}