;****************************************************************
;		      HuSound- Main Driver			*
;****************************************************************
; Version 1.3cz, November 10th, 2015
; Copyright (C) 2014-2015 Osman Celimli
;
; This software is provided 'as-is', without any express or implied
; warranty.  In no event will the authors be held liable for any damages
; arising from the use of this software.
;
; Permission is granted to anyone to use this software for any purpose,
; including commercial applications, and to alter it and redistribute it
; freely, subject to the following restrictions:
;
; 1. The origin of this software must not be misrepresented; you must not
;    claim that you wrote the original software. If you use this software
;    in a product, an acknowledgment in the product documentation would be
;    appreciated but is not required.
; 2. Altered source versions must be plainly marked as such, and must not be
;    misrepresented as being the original software.
; 3. This notice may not be removed or altered from any source distribution.
;----------------------------------------------------------------
;
; Library Parameters
; -------------------
; -HUSOUND_CHANNELS:
;	The number of channels HuSound will be driving, which is
; all six for most applications. Setting this to a value less than
; six will drop the higher channels (the ones with LFSRs) first.
;
; -HUSOUND_STACKLEN:
;	CALLs, LOOPs, and BREAKs require the use of a stack to
; keep track of where the last script decode address was. The stack
; length can be controlled by this parameter. A value of four is
; recommended for most users.
;
; -HUSOUND_REQLEN:
;	Controls the length of the sound effect dispatch request
; lists. Depending on how many sound effects you like to "stack,"
; this can range from 1 to 127. Eight is usually good enough.
;
;
; Notable Functions
; -----------------
; -HuSound_Init:
;	Initialize the HuSound driver and its subcomponents, should
; be called once during your system intialization routine.
;
; -HuSound_Main:
;	The main driver update routine, call once every VBL.
; -HuSound_Writeback:
;	The audio hardware writeback routine, this should be called
; as soon as possible after VBL occurs to ensure consistent timing.
;
; -HuSound_Pause:
; -HuSound_UnPause:
;	Pause or UnPause (Resume) audio playback, respectively.
; Both of these are protected against "double" pausing or resuming.
;
; -HuSound_StopAll:
;	Stop any music track or sound effect currently playing.
;
.section "HuSound_RefTables" force
HuSound_HasLFSR:
	.db $00,$00,$00,$00,$FF,$FF

HuSound_WaveOffs:
	.db $00,$20,$40,$60,$80,$A0

HuSound_DecOffs:
	.db $00,$04,$08,$0C,$10,$14,$1C
	.db $20,$24,$28,$2C,$30,$34,$3C

HuSound_LoopOffs:
	.db (HUSOUND_STACKLEN * 0),(HUSOUND_STACKLEN * 1)
	.db (HUSOUND_STACKLEN * 2),(HUSOUND_STACKLEN * 3)
	.db (HUSOUND_STACKLEN * 4),(HUSOUND_STACKLEN * 5)

HuSound_StackOffs:
	.db (HUSOUND_STACKLEN * 4 * 0),(HUSOUND_STACKLEN * 4 * 1)
	.db (HUSOUND_STACKLEN * 4 * 2),(HUSOUND_STACKLEN * 4 * 3)
	.db (HUSOUND_STACKLEN * 4 * 4),(HUSOUND_STACKLEN * 4 * 5)
.ends

;****************************************************************
; HuSound_Main:							*
;	The main driver tick, should be called once every VBL	*
; after calling HuSound_Writeback. The timing isn't critical,	*
; as the new audio state is only generated here, not written.	*
;****************************************************************
.section "HuSound_Main" force
HuSound_Main:
	LDA HuSound_Busy	; Check against double-entry, which is
	beq _safeStart		; more or less likely depending upon where
	rts			; the update routine is called from.
_safeStart
	LDA #$FF		; Set our busy flag and re-enable IRQs if
	STA HuSound_Busy	; it already hasn't been done. This way higher
	cli			; priority stuff (like raster fx) can happen
				; while we're updating audio.
	LDA HuSound_isPaused
	beq _beginDecode	; Of course, we shouldn't do anything if the
	STZ HuSound_Busy	; driver is in a paused state.
	rts
_beginDecode
	LDA MCG_LastMap		; Before we do anything, save the current
	PHA			; MCGenjin and MPR4 (OBJ_LO) states since
	TMA #MPR4		; we'll change them quite frequently during
	PHA			; our update routine.


_checkTrack			; ---- MUSIC DISPATCH ----
	LDA HuMusic_ReqPlay
	beq _procMusic

	STZ HuMusic_ReqPlay
	LDA HuMusic_ReqTrack
	jsr HuMusic_TrackPlay


_procMusic			; ---- MUSIC DECODE ----
	LDA HuMusic_isPlaying	; 
	beq _checkDispatch
	STZ HuMusic_ActiveAcc

	LDX #(HUSOUND_CHANNELS - 1)
_checkMusic
	LDA HuMusic_Priority,X	; Update each channel, checking to see if it's
	beq _skipMusic		; still active (priority > 0).

	jsr HuMusic_Decode
	INC HuMusic_ActiveAcc
_skipMusic
	DEX
	bpl _checkMusic

	LDA HuMusic_ActiveAcc	; If none of the music tracks are still playing,
	bne _checkDispatch	; clear the music playing flag. This way the user
	jsr HuMusic_Stop	; can actually check if "one shot" tracks finished.
	STZ HuMusic_isPlaying

_checkDispatch			; ---- SFX DISPATCH ----
	jsr HuSFX_SortReqs
	jsr HuSFX_Dispatch

_procSFX			; ---- SFX / INSTRUMENT DECODE ----
	LDX #(HUSOUND_CHANNELS - 1)
_checkSFX
	LDA HuSound_Priority,X
	beq _skipChannel

	jsr HuSFX_Decode
_skipChannel
	DEX
	bpl _checkSFX


	PLA			; Alright, we're done with the bank-altering
	TAM #MPR4		; updates, so put the mappers back as they were.
	PLA
	STA MCG_LastMap
	STA MCGENJIN_BANK

	LDX #(HUSOUND_CHANNELS - 1)
_updateStates			; Now that we've finished the music and SFX
	jsr HuSound_CalcNext	; decoding, we can finally calculate the new
	DEX			; state of the channels' frequency and volume
	bpl _updateStates	; registers. These will be written on the next
	STZ HuSound_Busy	; call to HuSound_Writeback.
	rts
.ends
;****************************************************************
; HuSound_Writeback:						*
;	Write the new state of the audio hardware to its regs	*
; in the HuC6280. This should be done as soon as possible after	*
; entering the VBL IRQ in order to get consistent timing.	*
;****************************************************************
.section "HuSound_Writeback" force
HuSound_Writeback:
	LDX #(HUSOUND_CHANNELS - 1)
	CLY

_handleChannel
	STX PSG_CHANNEL
	LDA HuSound_noWriteback,X
	beq _checkActive	; Check to see if writebacks are
	DEX			; disabled on this channel. If so,
	bpl _handleChannel	; we can just move on to the next one.
	rts

_checkActive			; ---- Activity ----
	LDA HuSound_Priority,X	; If writebacks are enabled, we're only
	bne _waveUpdate		; going to do anything if the channel is
				; active (priority other than zero). When
	LDA HuSound_Vol_Last,X	; idle, the channel will be forced silent.
	ORA #PSG_CONTROL_SAMPLE
	STA PSG_CONTROL		; <- Essential for reducing POPs when going
				; in and out of sampling mode.
	LDA.w HuSound_HasLFSR,X	
	beq _checkNext

	STZ PSG_LFSR
_checkNext
	DEX
	bpl _handleChannel
	rts

_waveUpdate			; ---- Waveform ----
	LDA HuSound_writeWave,X	; Load a new waveform only if we've been
	beq _freqUpdate		; asked, this will also silence the channel
				; until the panning and volume writes.
	LDY.w HuSound_WaveOffs,X
	LDA #PSG_CONTROL_WRITE
	STA PSG_CONTROL
	PHX
	LDX #31
_copyWave
	LDA HuSound_WaveMem,Y
	STA PSG_WAVEDAT
	INY
	DEX
	bpl _copyWave
	PLX
	STZ HuSound_writeWave,X

_freqUpdate			; ---- Frequency ----
	LDA HuSound_writeFreq,X	; Write the frequency to either the waveform
	beq _volUpdate		; frequency register or LFSR frequency register
				; if we've been requested to.
	LDA HuSound_WaveMode,X	; I'm not sure if writing to the frequency regs
	beq _waveFreq		; affects the divider state. But on some other
_noiseFreq			; denizens of the game console world it does.
	LDA HuSound_FreqFinal_Lo,X
	EOR #$FF
	AND #PSG_LFSR_FREQMASK
	ORA #PSG_LFSR_ENABLE
	STA PSG_LFSR		; <- This really wasn't thought through, Hudson.
	STZ HuSound_writeFreq,X
	bra _volUpdate
_waveFreq
	LDA HuSound_FreqFinal_Lo,X
	STA PSG_FREQ_LO
	LDA HuSound_FreqFinal_Hi,X
	STA PSG_FREQ_HI
	STZ HuSound_writeFreq,X

_volUpdate			; ---- Volume and Panning ----
	LDA HuSound_Panning,X	; These writes will always be done unless the
	STA PSG_PAN		; channel is disabled or its writebacks blocked.
	LDA HuSound_Vol_Lo,X
	AND #PSG_CONTROL_VOLMASK
	STA HuSound_Vol_Last,X
	ORA #PSG_CONTROL_WAVE
	STA PSG_CONTROL		; <- Always enabled after this write

_checkAllDone
	DEX
	bmi _done
	jmp _handleChannel
_done
	rts
.ends




;****************************************************************
; HuSound_CalcNext:						*
;	Calculate the next state for the channel indexed by X.	*
;****************************************************************
.section "HuSound_CalcNext" force
HuSound_CalcNext:
	LDA HuSound_Priority,X
	bne _isActive

	DEA
	STA HuSound_LastFreq_Lo,X
	STA HuSound_LastFreq_Hi,X
	rts

_isActive
	clc
	LDA HuSound_Vol_De,X	; ---- Volume ----
	ADC HuSound_VolAdj_De,X	; Should be recalculated independently
	STA HuSound_Vol_De,X	; of whether the channel is currently
	LDA HuSound_Vol_Lo,X	; occupied by an instrument or SFX.
	ADC HuSound_VolAdj_Lo,X
	STA HuSound_Vol_Lo,X


_checkType
	LDA HuSound_isInstrument,X
	beq _isSFX
_isInstrument			; ---- Frequency (Instrument) ----
	clc			; Full Base/Pitch/Offset Support.
	LDA HuSound_FreqPitch_De,X
	ADC HuSound_FreqPitchAdj_De,X
	STA HuSound_FreqPitch_De,X
	LDA HuSound_FreqPitch_Lo,X
	ADC HuSound_FreqPitchAdj_Lo,X
	STA HuSound_FreqPitch_Lo,X
	LDA HuSound_FreqPitch_Hi,X
	ADC HuSound_FreqPitchAdj_Hi,X
	STA HuSound_FreqPitch_Hi,X

	clc
	LDA HuSound_FreqOff_De,X
	ADC HuSound_FreqOffAdj_De,X
	STA HuSound_FreqOff_De,X
	LDA HuSound_FreqOff_Lo,X
	ADC HuSound_FreqOffAdj_Lo,X
	STA HuSound_FreqOff_Lo,X
	LDA HuSound_FreqOff_Hi,X
	ADC HuSound_FreqOffAdj_Hi,X
	STA HuSound_FreqOff_Hi,X

	clc
	LDA HuSound_FreqBase_De,X
	ADC HuSound_FreqPitch_De,X
	STA HuSound_FreqAcc_De
	LDA HuSound_FreqBase_Lo,X
	ADC HuSound_FreqPitch_Lo,X
	STA HuSound_FreqAcc_Lo
	LDA HuSound_FreqBase_Hi,X
	ADC HuSound_FreqPitch_Hi,X
	STA HuSound_FreqAcc_Hi

	clc
	LDA HuSound_FreqAcc_De
	ADC HuSound_FreqOff_De,X
	LDA HuSound_FreqAcc_Lo
	ADC HuSound_FreqOff_Lo,X
	STA HuSound_FreqFinal_Lo,X
	LDA HuSound_FreqAcc_Hi
	ADC HuSound_FreqOff_Hi,X
	STA HuSound_FreqFinal_Hi,X
	bra _checkFreqLo

_isSFX				; ---- Frequency (SFX) ----
	clc			; Sound Effects only use Offset/Adj.
	LDA HuSound_FreqOff_De,X
	ADC HuSound_FreqOffAdj_De,X
	STA HuSound_FreqOff_De,X
	LDA HuSound_FreqOff_Lo,X
	ADC HuSound_FreqOffAdj_Lo,X
	STA HuSound_FreqOff_Lo,X
	STA HuSound_FreqFinal_Lo,X
	LDA HuSound_FreqOff_Hi,X
	ADC HuSound_FreqOffAdj_Hi,X
	STA HuSound_FreqOff_Hi,X
	STA HuSound_FreqFinal_Hi,X

_checkFreqLo			; ---- Check for Frequency Change ----
	LDA HuSound_FreqFinal_Lo,X
	CMP HuSound_LastFreq_Lo,X
	beq _checkFreqHi
	LDA #$FF
	STA HuSound_writeFreq,X
	bra _cacheNewFreq
_checkFreqHi
	LDA HuSound_FreqFinal_Lo,X
	CMP HuSound_LastFreq_Lo,X
	beq _cacheNewFreq
	LDA #$FF
	STA HuSound_writeFreq,X
_cacheNewFreq
	LDA HuSound_FreqFinal_Lo,X
	STA HuSound_LastFreq_Lo,X
	LDA HuSound_FreqFinal_Hi,X
	STA HuSound_LastFreq_Hi,X
	rts
.ends




;****************************************************************
; HuSound_Pause:						*
;	Pause all audio decoding and silence output.		*
;****************************************************************
.section "HuSound_Pause" force
HuSound_Pause:
	LDA HuSound_isPaused
	beq _stopDecoding
	rts

_stopDecoding
	sei
	STZ TIMER_EN		; Disabling the timer and silencing
	STZ PSG_GLOPAN		; the global volume effectively halts
				; decoding (since we'll skip updates
				; if HuSound_isPaused is nonzero).
	LDA #$FF
	STA HuSound_isPaused	; We've officially shut up.
	cli
	rts
.ends
;****************************************************************
; HuSound_UnPause:						*
;	Resume audio playback from a paused state.		*
;****************************************************************
.section "HuSound_UnPause" force
HuSound_UnPause:
	LDA HuSound_isPaused
	bne _resumeDecoding
	rts

_resumeDecoding
	sei
	LDA #$FF
	STA TIMER_EN		; Re-enable the timer and turn
	STA PSG_GLOPAN		; global audio output back on.

	STZ HuSound_isPaused	; No longer paused!
	cli
	rts
.ends




;****************************************************************
; HuSound_StopAll:						*
;	Cease playback of any active music, sfx, or samples.	*
;****************************************************************
.section "HuSound_StopAll" force
HuSound_StopAll:
	jsr HuMusic_Stop
	jsr HuSFX_StopAll

	PHY
	LDY #(HUSOUND_CHANNELS - 1)
_killSamples
	jsr HuSample_Stop
	DEY
	bpl _killSamples
	rts
.ends




;****************************************************************
; HuSound_ReadByte:						*
;	Read the byte of data pointed to by HuSound_CurStream	*
; into HuSound_CurByte and A, then advance the pointer by one.	*
;****************************************************************
.section "HuSound_ReadByte" force
HuSound_ReadByte:
	PHY
	LDA HuSound_CurStream + MCMAP
	LDY SysRegion
	beq _setDirect
	TAY
	LDA MCG_SwapTable,Y
_setDirect
	PLY			; Set the MCGenjin Slot and MPR Bank,
	STA MCG_LastMap		; note that we could get interrupted in here!
	STA MCGENJIN_BANK	; Save the Map FIRST.
	LDA HuSound_CurStream + BANK
	TAM #MPR4

	LDA (HuSound_CurStream)
	STA HuSound_CurByte

	INC HuSound_CurStream + LO
	bne _noWrapSrc
	INC HuSound_CurStream + HI
	LDA HuSound_CurStream + HI
	CMP #>ADDR_OBJ_HI
	bne _noWrapSrc

	LDA #>ADDR_OBJ_LO
	STA HuSound_CurStream + HI
	INC HuSound_CurStream + BANK
	LDA HuSound_CurStream + BANK
	CMP #(BANK_ROM + MCBANK_PAGES)
	bne _noWrapSrc

	LDA #BANK_ROM
	STA HuSound_CurStream + BANK
	INC HuSound_CurStream + MCMAP
_noWrapSrc
	LDA HuSound_CurByte
	rts
.ends




;****************************************************************
; HuSound_Init:							*
;	Initialize the audio hardware along with the HuEffex,	*
; HuMusic, and HuSample libraries.				*
;****************************************************************
.section "HuSound_Init" force
HuSound_Init:
	LDA #$4C		; Ensure our VarJump is actually
	STA HuSound_VarJump	; a "JMP $abs"...

	jsr HuSFX_Init		; Set up HuSFX / SFX Database
	jsr HuMusic_Init	; Set up HuMusic / Track Database
	jsr HuSample_Init	; Set up HuSample / PCM Streaming

	LDA #$FF		; We're done setting up, so enable
	STA PSG_GLOPAN		; global audio output on both the L&R.
	rts
.ends
